@@ -305,7 +305,8 @@ g_socket_details_from_fd (GSocket *socket)
   int value;
   int errsv;
 #ifdef G_OS_WIN32
-  BOOL bool_val;
+  /* See bug #611756 */
+  BOOL bool_val = FALSE;
 #else
   int bool_val;
 #endif
@@ -388,7 +389,14 @@ g_socket_details_from_fd (GSocket *socket)
   if (getsockopt (fd, SOL_SOCKET, SO_KEEPALIVE,
 		  (void *)&bool_val, &optlen) == 0)
     {
+#ifndef G_OS_WIN32
+      /* Experimentation indicates that the SO_KEEPALIVE value is
+       * actually a char on Windows, even if documentation claims it
+       * to be a BOOL which is a typedef for int. So this g_assert()
+       * fails. See bug #611756.
+       */
       g_assert (optlen == sizeof bool_val);
+#endif
       socket->priv->keepalive = !!bool_val;
     }
   else
@@ -606,6 +614,12 @@ g_socket_finalize (GObject *object)
     g_socket_close (socket, NULL);
 
 #ifdef G_OS_WIN32
+  if (socket->priv->event != WSA_INVALID_EVENT)
+    {
+      WSACloseEvent (socket->priv->event);
+      socket->priv->event = WSA_INVALID_EVENT;
+    }
+
   g_assert (socket->priv->requested_conditions == NULL);
 #endif
 
@@ -1297,7 +1311,7 @@ g_socket_bind (GSocket         *socket,
  *
  * Returns: %TRUE if this socket can be used with IPv4.
  *
- * Since: 2.22.
+ * Since: 2.22
  **/
 gboolean
 g_socket_speaks_ipv4 (GSocket *socket)
@@ -1707,6 +1721,16 @@ g_socket_receive_from (GSocket         *socket,
 				   error);
 }
 
+/* Although we ignore SIGPIPE, gdb will still stop if the app receives
+ * one, which can be confusing and annoying. So if possible, we want
+ * to suppress the signal entirely.
+ */
+#ifdef MSG_NOSIGNAL
+#define G_SOCKET_DEFAULT_SEND_FLAGS MSG_NOSIGNAL
+#else
+#define G_SOCKET_DEFAULT_SEND_FLAGS 0
+#endif
+
 /**
  * g_socket_send:
  * @socket: a #GSocket
@@ -1759,7 +1783,7 @@ g_socket_send (GSocket       *socket,
 				    G_IO_OUT, cancellable, error))
 	return -1;
 
-      if ((ret = send (socket->priv->fd, buffer, size, 0)) < 0)
+      if ((ret = send (socket->priv->fd, buffer, size, G_SOCKET_DEFAULT_SEND_FLAGS)) < 0)
 	{
 	  int errsv = get_socket_errno ();
 
@@ -1984,14 +2008,6 @@ g_socket_close (GSocket  *socket,
       break;
     }
 
-#ifdef G_OS_WIN32
-  if (socket->priv->event != WSA_INVALID_EVENT)
-    {
-      WSACloseEvent (socket->priv->event);
-      socket->priv->event = WSA_INVALID_EVENT;
-    }
-#endif
-
   socket->priv->connected = FALSE;
   socket->priv->closed = TRUE;
 
@@ -2255,7 +2271,10 @@ winsock_finalize (GSource *source)
   g_object_unref (socket);
 
   if (winsock_source->cancellable)
-    g_object_unref (winsock_source->cancellable);
+    {
+      g_cancellable_release_fd (winsock_source->cancellable);
+      g_object_unref (winsock_source->cancellable);
+    }
 }
 
 static GSourceFuncs winsock_funcs =
@@ -2291,11 +2310,10 @@ winsock_source_new (GSocket      *socket,
   winsock_source->condition = condition;
   add_condition_watch (socket, &winsock_source->condition);
 
-  if (cancellable)
+  if (g_cancellable_make_pollfd (cancellable,
+                                 &winsock_source->cancel_pollfd))
     {
       winsock_source->cancellable = g_object_ref (cancellable);
-      g_cancellable_make_pollfd (cancellable,
-				 &winsock_source->cancel_pollfd);
       g_source_add_poll (source, &winsock_source->cancel_pollfd);
     }
 
@@ -2446,11 +2464,8 @@ g_socket_condition_wait (GSocket       *socket,
     num_events = 0;
     events[num_events++] = socket->priv->event;
 
-    if (cancellable)
-      {
-	g_cancellable_make_pollfd (cancellable, &cancel_fd);
-	events[num_events++] = (WSAEVENT)cancel_fd.fd;
-      }
+    if (g_cancellable_make_pollfd (cancellable, &cancel_fd))
+      events[num_events++] = (WSAEVENT)cancel_fd.fd;
 
     current_condition = update_condition (socket);
     while ((condition & current_condition) == 0)
@@ -2474,6 +2489,8 @@ g_socket_condition_wait (GSocket       *socket,
 	current_condition = update_condition (socket);
       }
     remove_condition_watch (socket, &condition);
+    if (num_events > 1)
+      g_cancellable_release_fd (cancellable);
 
     return (condition & current_condition) != 0;
   }
@@ -2487,15 +2504,15 @@ g_socket_condition_wait (GSocket       *socket,
     poll_fd[0].events = condition;
     num = 1;
 
-    if (cancellable)
-      {
-	g_cancellable_make_pollfd (cancellable, &poll_fd[1]);
-	num++;
-      }
+    if (g_cancellable_make_pollfd (cancellable, &poll_fd[1]))
+      num++;
 
     do
       result = g_poll (poll_fd, num, -1);
     while (result == -1 && get_socket_errno () == EINTR);
+    
+    if (num > 1)
+      g_cancellable_release_fd (cancellable);
 
     return cancellable == NULL ||
       !g_cancellable_set_error_if_cancelled (cancellable, error);
@@ -2618,6 +2635,11 @@ g_socket_send_message (GSocket                *socket,
 	if (!g_socket_address_to_native (address, msg.msg_name, msg.msg_namelen, error))
 	  return -1;
       }
+    else
+      {
+	msg.msg_name = NULL;
+	msg.msg_namelen = 0;
+      }
 
     /* iov */
     {
@@ -2680,7 +2702,7 @@ g_socket_send_message (GSocket                *socket,
 				      G_IO_OUT, cancellable, error))
 	  return -1;
 
-	result = sendmsg (socket->priv->fd, &msg, flags);
+	result = sendmsg (socket->priv->fd, &msg, flags | G_SOCKET_DEFAULT_SEND_FLAGS);
 	if (result < 0)
 	  {
 	    int errsv = get_socket_errno ();
@@ -3022,9 +3044,8 @@ g_socket_receive_message (GSocket                 *socket,
 	  if (index == allocated)
 	    {
 	      /* estimated 99% case: exactly 1 control message */
-	      allocated = MIN (allocated * 2, 1);
+	      allocated = MAX (allocated * 2, 1);
 	      my_messages = g_new (GSocketControlMessage *, (allocated + 1));
-	      allocated = 1;
 	    }
 
 	  my_messages[index++] = message;
