@@ -38,6 +38,9 @@
 #include <wchar.h>
 #include <direct.h>
 #include <io.h>
+#include <sys/utime.h>
+#else
+#include <utime.h>
 #endif
 
 #include "gstdio.h"
@@ -48,6 +51,10 @@
 #error Please port this to your operating system
 #endif
 
+#if defined (_MSC_VER) && !defined(_WIN64)
+#undef _wstat
+#define _wstat _wstat32
+#endif
 
 /**
  * g_access:
@@ -56,12 +63,16 @@
  *
  * A wrapper for the POSIX access() function. This function is used to
  * test a pathname for one or several of read, write or execute
- * permissions, or just existence. On Windows, the underlying access()
- * function in the C library only checks the READONLY attribute, and
- * does not look at the ACL at all. Software that needs to handle file
- * permissions on Windows more exactly should use the Win32 API.
+ * permissions, or just existence.
  *
- * See the C library manual for more details about access().
+ * On Windows, the file protection mechanism is not at all POSIX-like,
+ * and the underlying function in the C library only checks the
+ * FAT-style READONLY attribute, and does not look at the ACL of a
+ * file at all. This function is this in practise almost useless on
+ * Windows. Software that needs to handle file permissions on Windows
+ * more exactly should use the Win32 API.
+ *
+ * See your C library manual for more details about access().
  *
  * Returns: zero if the pathname refers to an existing file system
  * object that has all the tested permissions, or -1 otherwise or on
@@ -106,14 +117,15 @@ g_access (const gchar *filename,
  * @mode: as in chmod()
  *
  * A wrapper for the POSIX chmod() function. The chmod() function is
- * used to set the permissions of a file system object. Note that on
- * Windows the file protection mechanism is not at all POSIX-like, and
- * the underlying chmod() function in the C library just sets or
- * clears the READONLY attribute. It does not touch any ACL. Software
- * that needs to manage file permissions on Windows exactly should
- * use the Win32 API.
+ * used to set the permissions of a file system object.
+ * 
+ * On Windows the file protection mechanism is not at all POSIX-like,
+ * and the underlying chmod() function in the C library just sets or
+ * clears the FAT-style READONLY attribute. It does not touch any
+ * ACL. Software that needs to manage file permissions on Windows
+ * exactly should use the Win32 API.
  *
- * See the C library manual for more details about chmod().
+ * See your C library manual for more details about chmod().
  *
  * Returns: zero if the operation succeeded, -1 on error.
  * 
@@ -145,7 +157,6 @@ g_chmod (const gchar *filename,
   return chmod (filename, mode);
 #endif
 }
-
 /**
  * g_open:
  * @filename: a pathname in the GLib file name encoding (UTF-8 on Windows)
@@ -153,13 +164,23 @@ g_chmod (const gchar *filename,
  * @mode: as in open()
  *
  * A wrapper for the POSIX open() function. The open() function is
- * used to convert a pathname into a file descriptor. Note that on
- * POSIX systems file descriptors are implemented by the operating
- * system. On Windows, it's the C library that implements open() and
- * file descriptors. The actual Windows API for opening files is
- * something different.
+ * used to convert a pathname into a file descriptor.
  *
- * See the C library manual for more details about open().
+ * On POSIX systems file descriptors are implemented by the operating
+ * system. On Windows, it's the C library that implements open() and
+ * file descriptors. The actual Win32 API for opening files is quite
+ * different, see MSDN documentation for CreateFile(). The Win32 API
+ * uses file handles, which are more randomish integers, not small
+ * integers like file descriptors.
+ *
+ * Because file descriptors are specific to the C library on Windows,
+ * the file descriptor returned by this function makes sense only to
+ * functions in the same C library. Thus if the GLib-using code uses a
+ * different C library than GLib does, the file descriptor returned by
+ * this function cannot be passed to C library functions like write()
+ * or read().
+ *
+ * See your C library manual for more details about open().
  *
  * Returns: a new file descriptor, or -1 if an error occurred. The
  * return value can be used exactly like the return value from open().
@@ -201,12 +222,23 @@ g_open (const gchar *filename,
  *
  * A wrapper for the POSIX creat() function. The creat() function is
  * used to convert a pathname into a file descriptor, creating a file
- * if necessary. Note that on POSIX systems file descriptors are
- * implemented by the operating system. On Windows, it's the C library
- * that implements creat() and file descriptors. The actual Windows
- * API for opening files is something different.
+ * if necessary.
+
+ * On POSIX systems file descriptors are implemented by the operating
+ * system. On Windows, it's the C library that implements creat() and
+ * file descriptors. The actual Windows API for opening files is
+ * different, see MSDN documentation for CreateFile(). The Win32 API
+ * uses file handles, which are more randomish integers, not small
+ * integers like file descriptors.
  *
- * See the C library manual for more details about creat().
+ * Because file descriptors are specific to the C library on Windows,
+ * the file descriptor returned by this function makes sense only to
+ * functions in the same C library. Thus if the GLib-using code uses a
+ * different C library than GLib does, the file descriptor returned by
+ * this function cannot be passed to C library functions like write()
+ * or read().
+ *
+ * See your C library manual for more details about creat().
  *
  * Returns: a new file descriptor, or -1 if an error occurred. The
  * return value can be used exactly like the return value from creat().
@@ -249,9 +281,8 @@ g_creat (const gchar *filename,
  * renames a file, moving it between directories if required.
  * 
  * See your C library manual for more details about how rename() works
- * on your system. Note in particular that on Win9x it is not possible
- * to rename a file if a file with the new name already exists. Also
- * it is not possible in general on Windows to rename an open file.
+ * on your system. It is not possible in general on Windows to rename
+ * a file that is open to some process.
  *
  * Returns: 0 if the renaming succeeded, -1 if an error occurred
  * 
@@ -322,7 +353,7 @@ g_rename (const gchar *oldfilename,
  * attempts to create a directory with the given name and permissions.
  * The mode argument is ignored on Windows.
  * 
- * See the C library manual for more details about mkdir().
+ * See your C library manual for more details about mkdir().
  *
  * Returns: 0 if the directory was successfully created, -1 if an error 
  *    occurred
@@ -403,11 +434,11 @@ g_chdir (const gchar *path)
  *
  * A wrapper for the POSIX stat() function. The stat() function
  * returns information about a file. On Windows the stat() function in
- * the C library checks only the READONLY attribute and does not look
- * at the ACL at all. Thus the protection bits in the st_mode field
- * are a fabrication of little use.
+ * the C library checks only the FAT-style READONLY attribute and does
+ * not look at the ACL at all. Thus on Windows the protection bits in
+ * the st_mode field are a fabrication of little use.
  * 
- * See the C library manual for more details about stat().
+ * See your C library manual for more details about stat().
  *
  * Returns: 0 if the information was successfully retrieved, -1 if an error 
  *    occurred
@@ -415,8 +446,13 @@ g_chdir (const gchar *path)
  * Since: 2.6
  */
 int
-g_stat (const gchar *filename,
-	struct stat *buf)
+g_stat (const gchar           *filename,
+#ifdef G_OS_WIN32
+	struct _g_stat_struct *buf
+#else
+	struct stat           *buf
+#endif
+				  )
 {
 #ifdef G_OS_WIN32
   wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
@@ -437,7 +473,7 @@ g_stat (const gchar *filename,
       (!g_path_is_absolute (filename) || len > g_path_skip_root (filename) - filename))
     wfilename[len] = '\0';
 
-  retval = _wstat (wfilename, (struct _stat *) buf);
+  retval = _wstat (wfilename, (void*) buf);
   save_errno = errno;
 
   g_free (wfilename);
@@ -461,7 +497,7 @@ g_stat (const gchar *filename,
  * refers to. If the system does not support symbolic links g_lstat()
  * is identical to g_stat().
  * 
- * See the C library manual for more details about lstat().
+ * See your C library manual for more details about lstat().
  *
  * Returns: 0 if the information was successfully retrieved, -1 if an error 
  *    occurred
@@ -469,8 +505,13 @@ g_stat (const gchar *filename,
  * Since: 2.6
  */
 int
-g_lstat (const gchar *filename,
-	 struct stat *buf)
+g_lstat (const gchar           *filename,
+#ifdef G_OS_WIN32
+	 struct _g_stat_struct *buf
+#else
+	 struct stat           *buf
+#endif
+				   )
 {
 #ifdef HAVE_LSTAT
   /* This can't be Win32, so don't do the widechar dance. */
@@ -625,10 +666,18 @@ g_rmdir (const gchar *filename)
  * @mode: a string describing the mode in which the file should be 
  *   opened
  *
- * A wrapper for the POSIX fopen() function. The fopen() function opens
- * a file and associates a new stream with it. 
+ * A wrapper for the stdio fopen() function. The fopen() function
+ * opens a file and associates a new stream with it.
  * 
- * See the C library manual for more details about fopen().
+ * Because file descriptors are specific to the C library on Windows,
+ * and a file descriptor is partof the <type>FILE</type> struct, the
+ * <type>FILE</type> pointer returned by this function makes sense
+ * only to functions in the same C library. Thus if the GLib-using
+ * code uses a different C library than GLib does, the
+ * <type>FILE</type> pointer returned by this function cannot be
+ * passed to C library functions like fprintf() or fread().
+ *
+ * See your C library manual for more details about fopen().
  *
  * Returns: A <type>FILE</type> pointer if the file was successfully
  *    opened, or %NULL if an error occurred
@@ -683,7 +732,7 @@ g_fopen (const gchar *filename,
  * A wrapper for the POSIX freopen() function. The freopen() function
  * opens a file and associates it with an existing stream.
  * 
- * See the C library manual for more details about freopen().
+ * See your C library manual for more details about freopen().
  *
  * Returns: A <type>FILE</type> pointer if the file was successfully
  *    opened, or %NULL if an error occurred.
@@ -729,5 +778,48 @@ g_freopen (const gchar *filename,
 #endif
 }
 
+/**
+ * g_utime:
+ * @filename: a pathname in the GLib file name encoding (UTF-8 on Windows)
+ * @utb: a pointer to a struct utimbuf.
+ *
+ * A wrapper for the POSIX utime() function. The utime() function
+ * sets the access and modification timestamps of a file.
+ * 
+ * See your C library manual for more details about how utime() works
+ * on your system.
+ *
+ * Returns: 0 if the operation was successful, -1 if an error 
+ *    occurred
+ * 
+ * Since: 2.18
+ */
+int
+g_utime (const gchar    *filename,
+	 struct utimbuf *utb)
+{
+#ifdef G_OS_WIN32
+  wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
+  int retval;
+  int save_errno;
+
+  if (wfilename == NULL)
+    {
+      errno = EINVAL;
+      return -1;
+    }
+  
+  retval = _wutime (wfilename, (struct _utimbuf*) utb);
+  save_errno = errno;
+
+  g_free (wfilename);
+
+  errno = save_errno;
+  return retval;
+#else
+  return utime (filename, utb);
+#endif
+}
+
 #define __G_STDIO_C__
 #include "galiasdef.c"
